# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import oslotest.base

from designate.cmd import status


class StatusTestCase(oslotest.base.BaseTestCase):
    @mock.patch('designate.cmd.status.upgradecheck.main')
    @mock.patch('designate.cmd.status.utils.find_config')
    def test_main(self, mock_find_config, mock_upgradecheck_main):
        mock_find_config.return_value = ['/etc/designate/designate.conf']
        mock_upgradecheck_main.return_value = 0

        result = status.main()

        self.assertEqual(0, result)
        mock_find_config.assert_called_once_with('designate.conf')
        mock_upgradecheck_main.assert_called_once()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_utils import timeutils

from designate.manage import service
from designate import objects
from designate.tests import base_fixtures
import designate.tests.functional

CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class ManageServiceTestCase(designate.tests.functional.TestCase):
    def setUp(self):
        super().setUp()
        self.stdlog = base_fixtures.StandardLogging()
        self.useFixture(self.stdlog)
        self.command = service.ServiceCommands()

    @mock.patch.object(service, 'LOG')
    def test_service_clean(self, m_log):
        values = self.get_service_status_fixture()
        service_status = objects.ServiceStatus.from_dict(values)
        service_status.heartbeated_at = timeutils.datetime.datetime(
            2024, 5, 8, 14, 46, 38, 314323)
        self.storage.create_service_status(
            self.admin_context, service_status)

        values_new = self.get_service_status_fixture(fixture=1)
        service_status_new = objects.ServiceStatus.from_dict(values_new)
        self.storage.create_service_status(
            self.admin_context, service_status_new)
        self.assertEqual(
            len(self.storage.find_service_statuses(self.admin_context)), 2
        )
        self.command.clean()
        self.assertIn(
            service_status.service_name,
            m_log.warning.call_args_list[0].args[1]['service_name']
        )
        statuses = self.storage.find_service_statuses(self.admin_context)
        self.assertEqual(len(statuses), 1)

        # Make sure the remaining service is not the one who expired
        self.assertEqual(service_status_new.service_name,
                         statuses[0].service_name)
        self.assertNotEqual(service_status.service_name,
                            statuses[0].service_name)

    @mock.patch.object(service, 'LOG')
    def test_service_clean_no_dead_service(self, m_log):
        values = self.get_service_status_fixture()
        service_status = objects.ServiceStatus.from_dict(values)
        self.storage.create_service_status(
            self.admin_context, service_status)

        values_new = self.get_service_status_fixture(fixture=1)
        service_status_new = objects.ServiceStatus.from_dict(values_new)
        self.storage.create_service_status(
            self.admin_context, service_status_new)
        self.assertEqual(
            len(self.storage.find_service_statuses(self.admin_context)), 2
        )
        self.command.clean()
        self.assertEqual(
            len(m_log.warning.call_args_list), 0
        )
        statuses = self.storage.find_service_statuses(self.admin_context)
        self.assertEqual(len(statuses), 2)

    @mock.patch.object(service, 'LOG')
    def test_service_clean_message_timeout(self, m_log):
        self.command.storage.find_service_statuses = mock.Mock(
            side_effect=messaging.exceptions.MessagingTimeout
        )
        self.assertRaises(SystemExit, self.command.clean)
        self.assertEqual(
            len(m_log.critical.call_args_list), 1
        )



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Endre Karlson <endre.karlson@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from dns import reversename
from oslo_log import log as logging

import designate.conf
from designate import exceptions
from designate.plugin import DriverPlugin


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class NetworkAPI(DriverPlugin):
    """
    Base API
    """
    __plugin_ns__ = 'designate.network_api'
    __plugin_type__ = 'network_api'

    def _endpoints(self, service_catalog=None, service_type=None,
                   endpoint_type='publicURL', config_section=None,
                   region=None):
        configured_endpoints = self.get_configured_endpoints(config_section)
        if configured_endpoints:
            endpoints = self.endpoints_from_config(
                configured_endpoints,
                region=region,
            )
        elif service_catalog:
            endpoints = self.endpoints_from_catalog(
                service_catalog, service_type, endpoint_type,
                region=region,
            )
        else:
            raise exceptions.ConfigurationError(
                'No service_catalog and no configured endpoints'
            )

        LOG.debug('Returning endpoints: %s', endpoints)
        return endpoints

    @staticmethod
    def endpoints_from_config(configured_endpoints, region=None):
        """
        Return the endpoints for the given service from the configuration.

        return [('http://endpoint', 'region')]
        """
        endpoints = []
        for endpoint_data in configured_endpoints:
            if not endpoint_data:
                continue
            endpoint_region, endpoint = endpoint_data.split('|')
            if region and endpoint_region != region:
                continue
            endpoints.append((endpoint, endpoint_region))
        if not endpoints:
            raise exceptions.ConfigurationError(
                'Endpoints are not correctly configured'
            )
        return endpoints

    @staticmethod
    def endpoints_from_catalog(service_catalog, service_type, endpoint_type,
                               region=None):
        """
        Return the endpoints for the given service from the context's sc
        or lookup towards the configured keystone.

        return [('http://endpoint', 'region')]
        """
        endpoints = []
        for svc in service_catalog:
            if svc['type'] != service_type:
                continue
            for endpoint_data in svc['endpoints']:
                if endpoint_type not in endpoint_data:
                    continue
                endpoint = endpoint_data[endpoint_type]
                endpoint_region = endpoint_data['region']
                if region and endpoint_region != region:
                    continue
                endpoints.append((endpoint, endpoint_region))
        if not endpoints:
            raise exceptions.NetworkEndpointNotFound()
        return endpoints

    @staticmethod
    def get_configured_endpoints(config_section):
        """
        Returns endpoints from a specific section in the configuration.

        return ['region|http://endpoint']
        """
        if not config_section:
            return None
        cfg_group = CONF[config_section]
        return cfg_group.endpoints

    def list_floatingips(self, context, region=None):
        """
        List Floating IPs.

        Should return something like:

        [{
            'address': '<ip address'>,
            'region': '<region where this belongs>',
            'id': '<id of the FIP>'
        }]
        """
        raise NotImplementedError

    @staticmethod
    def address_zone(address):
        """
        Get the zone a address belongs to.
        """
        parts = reversed(address.split('.')[:-1])
        return '%s.in-addr.arpa.' % ".".join(parts)

    @staticmethod
    def address_name(address):
        """
        Get the name for the address
        """
        name = reversename.from_address(address).to_text()
        return name



# Copyright 2025 Cloudification GmbH
#
# Author: cloudification <contact@cloudification.io>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add SVCB and https record types

Revision ID: f828412479ee
Revises: 9099de8ae11c
Create Date: 2025-05-09 15:23:12.256620

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f828412479ee'
down_revision = '9099de8ae11c'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Check if the equivalent legacy migration has already run

    RECORD_TYPES = ['A', 'AAAA', 'CNAME', 'MX', 'SRV', 'TXT', 'SPF', 'NS',
                    'PTR', 'SSHFP', 'SOA', 'NAPTR', 'CAA', 'CERT', 'HTTPS',
                    'SVCB']

    with op.batch_alter_table('recordsets') as batch_op:
        batch_op.alter_column('type', type_=sa.Enum(name='record_types',
                                                    *RECORD_TYPES),
                              existing_type=sa.Enum, existing_nullable=False)



# Copyright 2025 Cloudification GmbH
#
# Author: cloudification <contact@cloudification.io>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64

from oslo_log import log as logging
import oslotest.base

from designate.exceptions import InvalidObject
from designate import objects

LOG = logging.getLogger(__name__)


class SVCBTest(oslotest.base.BaseTestCase):

    def test_svcb_record(self):
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data='1 sample.example.org.'
                         ' alpn=http/1.1,h2 dohpath=/dns-query{?dns}'
                ),
            ])
        )

        recordset.validate()

    def test_alpn_without_dohpath(self):
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data='1 sample.example.org.'
                         ' alpn=http/1.1,h2'
                ),
            ])
        )

        recordset.validate()

    def test_parse_svcb(self):
        svcb_record = objects.SVCB()
        ipv61 = '2001:db8:3333:4444:5555:6666:7777:8888'
        ipv6hint = f'{ipv61},2001:db8:3333:4444:cccc:dddd:eeee:ffff'
        ipv4hint = "1.2.3.4,9.8.7.6"
        alpn = "h3,h2,http/1.1"
        port = "888"
        target = "sample.example.org."
        dohpath = '/dns-query{?dns}'
        svcb_record.from_string(f'1 {target} alpn={alpn}'
                                f' ipv4hint={ipv4hint} '
                                f'ipv6hint={ipv6hint} port={port}'
                                f' dohpath={dohpath}'
                                )

        self.assertEqual(1, svcb_record.priority)
        self.assertEqual(target, svcb_record.target)
        self.assertEqual(f"alpn={alpn}", svcb_record.alpn)
        self.assertEqual(f"ipv6hint={ipv6hint}", svcb_record.ipv6hint)
        self.assertEqual(f"ipv4hint={ipv4hint}", svcb_record.ipv4hint)
        self.assertEqual(f"port={port}", svcb_record.port)
        self.assertEqual(f"dohpath={dohpath}", svcb_record.dohpath)

    def test_no_valid_alpn(self):
        doh = "dohpath=/dns-query{?dns}"
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(data=f'1 sample.example.org. alpn=foo {doh}'),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_non_valid_port(self):
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(data='1 sample.example.org. port=foo'),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_ech(self):
        svcb_record = objects.SVCB()
        ipv4hint = "1.2.3.4"
        alpn = "h2"
        target = "sample.example.org."
        ech = base64.b64encode(target.encode()).decode()
        doh = '/dns-query{?dns}'
        svcb_record.from_string(f'1 {target} alpn={alpn}'
                                f' ipv4hint={ipv4hint}'
                                f' dohpath={doh}'
                                f' ech={ech}'
                                )

        self.assertEqual(1, svcb_record.priority)
        self.assertEqual(target, svcb_record.target)
        self.assertEqual(f"alpn={alpn}", svcb_record.alpn)
        self.assertEqual(f"ipv4hint={ipv4hint}", svcb_record.ipv4hint)
        self.assertEqual(f"ech={ech}", svcb_record.ech)

    def test_echconfig(self):
        alpn = "h3"
        target = "sample.example.org."
        ech = base64.b64encode(target.encode()).decode()

        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data=f'1 sample.example.org.'
                         f' alpn={alpn} echconfig={ech}'),
            ])
        )
        self.assertRaises(InvalidObject, recordset.validate)

    def test_ech_validate(self):
        svcb_record = objects.SVCB()
        ipv4hint = "1.2.3.4"
        alpn = "h3"
        target = "sample.example.org."
        doh = '/dns-query{?dns}'
        ech = base64.b64encode(target.encode()).decode()
        svcb_record.from_string(f'1 {target} alpn={alpn}'
                                f' ipv4hint={ipv4hint}'
                                f' ech={ech}'
                                f' dohpath={doh}'
                                )

        self.assertEqual(1, svcb_record.priority)
        self.assertEqual(target, svcb_record.target)
        self.assertEqual(f"alpn={alpn}", svcb_record.alpn)
        self.assertEqual(f"ipv4hint={ipv4hint}", svcb_record.ipv4hint)
        self.assertEqual(f"ech={ech}", svcb_record.ech)

    def test_non_valid_ipv4hint(self):
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data='1 sample.example.org. ipv4hint=foo,192.168.1.2'
                ),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_escape_string_ech(self):
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(data='1 sample.example.org.'
                                    ' ech=foo//'),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_non_base64_ech(self):
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(data='1 sample.example.org.'
                                    ' ech=textstring'),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_non_valid_ipv6hint(self):
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(data='1 sample.example.org. ipv6hint=foo'),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_unsupported_param(self):
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data='1 sample.example.org. bazz=foo'
                ),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_doh(self):
        doh = '/dns-query{?dns}'
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data=f'1 sample.example.org. dohpath={doh}'
                ),
            ])
        )
        recordset.validate()

    def test_doh_not_relative(self):
        doh = 'dns-query{?dns}'
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data=f'1 sample.example.org. dohpath={doh}'
                ),
            ])
        )
        self.assertRaises(InvalidObject, recordset.validate)

    def test_doh_not_missed_dns(self):
        doh = '/dns-query{?test}'
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data=f'1 sample.example.org. dohpath={doh}'
                ),
            ])
        )
        self.assertRaises(InvalidObject, recordset.validate)

    def test_doh_short(self):
        doh = '/{dns}'
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data=f'1 sample.example.org. dohpath={doh}'
                ),
            ])
        )
        self.assertRaises(InvalidObject, recordset.validate)

    def test_no_doh_h3_alpn(self):
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data='1 sample.example.org. alpn=h3'
                ),
            ])
        )
        recordset.validate()

    def test_no_doh_h2_alpn(self):
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data='1 sample.example.org. alpn=h2'
                ),
            ])
        )
        recordset.validate()

    def test_no_doh_complex_alpn(self):
        recordset = objects.RecordSet(
            name='example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data='1 sample.example.org. alpn=h3,h2,http/1.1'
                ),
            ])
        )
        recordset.validate()

    def test_escape_in_ech(self):
        recordset = objects.RecordSet(
            name='example.org.', type='HTTPS',
            records=objects.RecordList(objects=[
                objects.Record(
                    data='1 sample.example.org. port=8888 ech=foo//'
                ),
            ])
        )
        self.assertRaises(InvalidObject, recordset.validate)

    def test_dns_no_doh_h3_alpn(self):
        recordset = objects.RecordSet(
            name='_dns.example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data='1 sample.example.org. alpn=h3'
                ),
            ])
        )
        self.assertRaises(InvalidObject, recordset.validate)

    def test_dns_no_doh_h2_alpn(self):
        recordset = objects.RecordSet(
            name='_dns.example.org.', type='SVCB',
            records=objects.RecordList(objects=[
                objects.Record(
                    data='1 sample.example.org. alpn=h2'
                ),
            ])
        )
        self.assertRaises(InvalidObject, recordset.validate)



# Copyright (c) 2025 VEXXHOST, Inc.
# SPDX-License-Identifier: Apache-2.0
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_utils import timeutils

import designate.conf
from designate.manage import base
from designate import rpc
from designate import storage


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class ServiceCommands(base.Commands):
    def __init__(self):
        super().__init__()
        self.heartbeat_interval = None
        self.storage = storage.get_storage()

    def clean(self):
        rpc.init(CONF)
        self.heartbeat_interval = CONF.heartbeat_emitter.heartbeat_interval
        LOG.info("Start cleaning dead services.")
        try:
            statuses = self.storage.find_service_statuses(self.context)
            for status in statuses:
                if status.heartbeated_at:
                    # Clean stale servcie if it pass 2*(heartbeat_interval)
                    check_interval = (
                        timeutils.utcnow() - status.heartbeated_at
                    ).total_seconds()
                    if check_interval > 2 * self.heartbeat_interval:
                        LOG.warning("Found dead service for delete: "
                                    "%(service_name)s. "
                                    "Last service heartbeat time is "
                                    "%(check_interval)s seconds ago.",
                                    {
                                        'service_name': status.service_name,
                                        'check_interval': check_interval
                                    }
                                    )
                        self.storage.delete_service_status(
                            self.context, status)
        except messaging.exceptions.MessagingTimeout:
            LOG.critical(
                'No response received from designate-central. '
                'Check it is running, and retry'
            )
            raise SystemExit(1)

        LOG.info("Job finished.")



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# NOTE: This module provides WSGIService which uses oslo_service.wsgi and
# oslo_service.sslutils. These modules will be removed in oslo.service 2026.2
# as they are built on eventlet. WSGIService and the designate-api service are
# deprecated. Deploy Designate API via uwsgi instead. See the DevStack
# configuration for an example of uwsgi deployment.

from debtcollector import removals
from oslo_log import log as logging
from oslo_utils import netutils
from paste import deploy

import designate.conf
from designate import exceptions
from designate import service
from designate import utils


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)

removals.removed_module(
    __name__,
    removal_version="2027.1",
    message=(
        "The designate.api.service module is deprecated. WSGIService uses "
        "oslo_service.wsgi and oslo_service.sslutils which will be removed "
        "in oslo.service 2026.2. Deploy Designate API via uwsgi instead. "
        "See the DevStack configuration for an example of uwsgi deployment."
    ),
)


class WSGIService(service.Service):
    def __init__(self, app, name, listen, max_url_len=None):
        from oslo_service import sslutils
        from oslo_service import wsgi

        super().__init__(name)
        self.app = app
        self.name = name

        self.listen = listen

        self.servers = []

        for address in self.listen:
            host, port = netutils.parse_host_port(address)
            server = wsgi.Server(
                CONF, name, app,
                host=host,
                port=port,
                pool_size=CONF['service:api'].threads,
                backlog=CONF.backlog,
                use_ssl=sslutils.is_enabled(CONF),
                max_url_len=max_url_len
            )

            self.servers.append(server)

    def start(self):
        for server in self.servers:
            server.start()
        super().start()

    def stop(self, graceful=True):
        for server in self.servers:
            server.stop()
        super().stop(graceful)

    def wait(self):
        for server in self.servers:
            server.wait()
        super().wait()


class Service(WSGIService):
    def __init__(self):
        super().__init__(
            self.wsgi_application,
            self.service_name,
            CONF['service:api'].listen,
        )

    def start(self):
        super().start()

    def stop(self, graceful=True):
        super().stop(graceful)

    @property
    def service_name(self):
        return 'api'

    @property
    def wsgi_application(self):
        api_paste_config = CONF['service:api'].api_paste_config
        config_paths = utils.find_config(api_paste_config)

        if not config_paths:
            raise exceptions.ConfigurationError(
                'Unable to determine appropriate api-paste-config file'
            )

        LOG.info('Using api-paste-config found at: %s', config_paths[0])

        return deploy.loadapp("config:%s" % config_paths[0], name='osapi_dns')



# Copyright 2025 Cloudification GmbH
#
# Author: cloudification <contact@cloudification.io>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging
import oslotest.base

from designate.exceptions import InvalidObject
from designate import objects

LOG = logging.getLogger(__name__)


class HTTPSTest(oslotest.base.BaseTestCase):

    def test_https_record(self):
        recordset = objects.RecordSet(
            name='example.org.', type='HTTPS',
            records=objects.RecordList(objects=[
                objects.Record(data='1 sample.example.org.'
                                    ' alpn=http/1.1,h2'),
            ])
        )

        recordset.validate()

    def test_parse_https(self):
        https_record = objects.HTTPS()
        ipv61 = '2001:db8:3333:4444:5555:6666:7777:8888'
        ipv6hint = f"{ipv61},2001:db8:3333:4444:cccc:dddd:eeee:ffff"
        ipv4hint = "1.2.3.4,9.8.7.6"
        alpn = "h3,h2,http/1.1"
        port = "888"
        target = 'sample.example.org.'
        https_record.from_string(f'1 {target}'
                                 f' alpn={alpn}'
                                 f' ipv4hint={ipv4hint}'
                                 f' ipv6hint={ipv6hint}'
                                 f' port={port}'
                                 )
        self.assertEqual(1, https_record.priority)
        self.assertEqual(target, https_record.target)
        self.assertEqual(f"alpn={alpn}", https_record.alpn)
        self.assertEqual(f"ipv6hint={ipv6hint}", https_record.ipv6hint)
        self.assertEqual(f"ipv4hint={ipv4hint}", https_record.ipv4hint)
        self.assertEqual(f"port={port}", https_record.port)

    def test_no_valid_alpn(self):
        recordset = objects.RecordSet(
            name='example.org.', type='HTTPS',
            records=objects.RecordList(objects=[
                objects.Record(data='1 sample.example.org. alpn=foo'),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_non_valid_port(self):
        recordset = objects.RecordSet(
            name='example.org.', type='HTTPS',
            records=objects.RecordList(objects=[
                objects.Record(data='1 sample.example.org.'
                                    ' port=foo'),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_non_valid_ech(self):
        recordset = objects.RecordSet(
            name='example.org.', type='HTTPS',
            records=objects.RecordList(objects=[
                objects.Record(data='1 sample.example.org.'
                                    ' ech=foo//'),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_non_valid_ipv4hint(self):
        recordset = objects.RecordSet(
            name='example.org.', type='HTTPS',
            records=objects.RecordList(objects=[
                objects.Record(data='1 sample.example.org.'
                                    ' ipv4hint=foo,192.168.1.2'),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_non_valid_ipv6hint(self):
        recordset = objects.RecordSet(
            name='example.org.', type='HTTPS',
            records=objects.RecordList(objects=[
                objects.Record(data='1 sample.example.org.'
                                    ' ipv6hint=foo'),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_no_default_alpn(self):
        recordset = objects.RecordSet(
            name='example.org.', type='HTTPS',
            records=objects.RecordList(objects=[
                objects.Record(data='1 sample.example.org. '
                                    'alpn=http/1.1 '
                                    'no-default-alpn port=8000'),
            ])
        )

        recordset.validate()

    def test_no_default_alpn_no_alpn_param(self):
        recordset = objects.RecordSet(
            name='example.org.', type='HTTPS',
            records=objects.RecordList(objects=[
                objects.Record(data='1 sample.example.org. '
                                    'no-default-alpn port=8000'),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_dohpath_invalid(self):
        query = "/dns-query{?dns}"
        data = f'1 doh.example. alpn=h2 dohpath={query}'
        recordset = objects.RecordSet(
            name='example.org.', type='HTTPS',
            records=objects.RecordList(objects=[
                objects.Record(data=data),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)

    def test_mandatory(self):
        ipv6 = "2001:db8:3333:4444:5555:6666:7777:8888"
        recordset = objects.RecordSet(
            name='example.org.', type='HTTPS',
            records=objects.RecordList(objects=[
                objects.Record(
                    data='1 sample.example.org.'
                         ' ipv4hint=192.168.1.2'
                         f' ipv6hint={ipv6}'
                         ' alpn=h2'
                         ' mandatory=alpn,ipv4hint'),
            ])
        )

        recordset.validate()

    def test_mandatory_negative(self):
        ipv6hint = "2001:db8:3333:4444:5555:6666:7777:8888"
        recordset = objects.RecordSet(
            name='example.org.', type='HTTPS',
            records=objects.RecordList(objects=[
                objects.Record(
                    data=f'1 sample.example.org.'
                         f' ipv6hint={ipv6hint}'
                         f' mandatory=alpn,ipv4hint'),
            ])
        )

        self.assertRaises(InvalidObject, recordset.validate)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import oslotest.base


class WSGIApiTestCase(oslotest.base.BaseTestCase):
    @mock.patch('designate.api.wsgi.init_application')
    def test_wsgi_api_application(self, mock_init_app):
        mock_init_app.return_value = mock.Mock()

        # Import the module to test the module-level code
        import designate.wsgi.api

        # Verify the application was initialized
        self.assertIsNotNone(designate.wsgi.api.application)
        mock_init_app.assert_called_once()



# Copyright 2016 Hewlett Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging
from stevedore import named

import designate.conf
from designate import exceptions


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class Scheduler:
    """Scheduler that schedules zones based on the filters provided on the zone
    and other inputs.

    :raises: NoFiltersConfigured
    """
    filters = []
    """The list of filters enabled on this scheduler"""

    def __init__(self, storage):
        enabled_filters = CONF['service:central'].scheduler_filters
        self.filters = list()
        self.storage = storage

        if not enabled_filters:
            raise exceptions.NoFiltersConfigured(
                'There are no scheduling filters configured'
            )

        extensions = named.NamedExtensionManager(
            namespace='designate.scheduler.filters',
            names=enabled_filters,
            name_order=True,
        )

        for extension in extensions:
            plugin = extension.plugin(storage=self.storage)
            LOG.info('Loaded Scheduler Filter: %s', plugin.name)
            self.filters.append(plugin)

    def schedule_zone(self, context, zone):
        """Get a pool to create the new zone in.

        :param context: :class:`designate.context.DesignateContext` - Context
            Object from request
        :param zone: :class:`designate.objects.zone.Zone` - Zone to be created
        :return: string -- ID of pool to schedule the zone to.
        :raises: MultiplePoolsFound, NoValidPoolFound
        """
        pools = self.storage.find_pools(context)

        if not self.filters:
            raise exceptions.NoFiltersConfigured(
                'There are no scheduling filters configured'
            )

        for plugin in self.filters:
            LOG.debug(
                'Running %s filter with %d pools', plugin.name, len(pools)
            )
            pools = plugin.filter(context, pools, zone)
            LOG.debug(
                '%d candidate pools remaining after %s filter',
                len(pools), plugin.name
            )

        if len(pools) > 1:
            raise exceptions.MultiplePoolsFound()
        if not pools:
            raise exceptions.NoValidPoolFound(
                'There are no pools that matched your request'
            )
        return pools[0].id



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# Copyright 2011 OpenStack Foundation
# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import errno
import socket
import struct
import threading

from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_service import service
from oslo_utils import netutils

from designate.common.decorators import rpc as rpc_decorator
from designate.common import profiler
import designate.conf
from designate.i18n import _
from designate import policy
from designate import rpc
from designate import utils
from designate import version

CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class Service(service.Service):
    def __init__(self, name, threads=None):
        threads = threads or 1000
        super().__init__(threads)
        self.name = name
        self.host = CONF.host

        policy.init()

        if not rpc.initialized():
            rpc.init(CONF)

        profiler.setup_profiler((''.join(('designate-', self.name))),
                                self.host)

    def start(self):
        LOG.info('Starting %(name)s service (version: %(version)s)',
                 {
                     'name': self.name,
                     'version': version.version_info.version_string()
                 })
        super().start()

    def stop(self, graceful=True):
        LOG.info('Stopping %(name)s service', {'name': self.name})
        super().stop(graceful)


class RPCService(Service):
    def __init__(self, name, rpc_topic, threads=None):
        super().__init__(name, threads)
        LOG.debug("Creating RPC Server on topic '%s' for %s",
                  rpc_topic, self.name)

        self.endpoints = [self]
        self.exception_thread_local = rpc_decorator.ExceptionThreadLocal()
        self.notifier = None
        self.rpc_server = None
        self.rpc_topic = rpc_topic
        self._rpc_initialized = False

    def init_host(self):
        """Initialize RPC server after service launch.

        This method is called after the service is created but before
        it starts handling requests. This ensures RPC initialization
        happens in the correct threading context.
        """
        if self._rpc_initialized:
            return

        LOG.info('Initializing RPC server for %s', self.name)
        target = messaging.Target(topic=self.rpc_topic, server=self.host)
        self.rpc_server = rpc.get_server(target, self.endpoints)
        self.rpc_server.start()
        self.notifier = rpc.get_notifier(self.name)
        self._rpc_initialized = True

    def start(self):
        super().start()
        if not self._rpc_initialized:
            self.init_host()

    def stop(self, graceful=True):
        if self.rpc_server:
            self.rpc_server.stop()
        super().stop(graceful)

    def wait(self):
        super().wait()


class DNSService:
    _TCP_RECV_MAX_SIZE = 65535

    def __init__(self, app, tg, listen, tcp_backlog, tcp_keepidle,
                 tcp_recv_timeout):
        self._running = threading.Event()
        self.app = app
        self.tg = tg
        self.tcp_backlog = tcp_backlog
        self.tcp_keepidle = tcp_keepidle
        self.tcp_recv_timeout = tcp_recv_timeout
        self.listen = listen

        self._dns_socks_tcp = []
        self._dns_socks_udp = []

    def start(self):
        self._running.set()

        addresses = map(
            netutils.parse_host_port,
            set(self.listen)
        )

        for address in addresses:
            self._start(address[0], address[1])

    def _start(self, host, port):
        sock_tcp = utils.bind_tcp(
            host, port, self.tcp_backlog, self.tcp_keepidle
        )
        sock_udp = utils.bind_udp(
            host, port
        )

        self._dns_socks_tcp.append(sock_tcp)
        self._dns_socks_udp.append(sock_udp)

        self.tg.add_thread(self._dns_handle_tcp, sock_tcp)
        self.tg.add_thread(self._dns_handle_udp, sock_udp)

    def stop(self):
        self._running.clear()

        for sock_tcp in self._dns_socks_tcp:
            sock_tcp.close()

        for sock_udp in self._dns_socks_udp:
            sock_udp.close()

    def _dns_handle_tcp(self, sock_tcp):
        LOG.info('_handle_tcp thread started')

        client = None
        while self._running.is_set():
            addr = None
            try:
                # handle a new TCP connection
                client, addr = sock_tcp.accept()

                if self.tcp_recv_timeout:
                    client.settimeout(self.tcp_recv_timeout)

                LOG.debug(
                    'Handling TCP Request from: %(host)s:%(port)d',
                    {
                        'host': addr[0],
                        'port': addr[1]
                    }
                )
                if len(addr) == 4:
                    LOG.debug(
                        'Flow info: %(host)s scope: %(port)d',
                        {
                            'host': addr[2],
                            'port': addr[3]
                        }
                    )

                # Dispatch a thread to handle the connection
                self.tg.add_thread(self._dns_handle_tcp_conn, addr, client)

            # NOTE: Any uncaught exceptions will result in the main loop
            # ending unexpectedly. Ensure proper ordering of blocks, and
            # ensure no exceptions are generated from within.
            except socket.timeout:
                pass
            except OSError as e:
                if client:
                    client.close()
                errname = errno.errorcode[e.args[0]]
                addr = addr or (None, 0)
                LOG.warning(
                    'Socket error %(err)s from: %(host)s:%(port)d',
                    {
                        'host': addr[0],
                        'port': addr[1],
                        'err': errname
                    }
                )
            except Exception:
                if client:
                    client.close()
                addr = addr or (None, 0)
                LOG.exception(
                    'Unknown exception handling TCP request from: '
                    '%(host)s:%(port)d',
                    {
                        'host': addr[0],
                        'port': addr[1]
                    }
                )

    def _dns_handle_tcp_conn(self, addr, client):
        """
        Handle a DNS Query over TCP. Multiple queries can be pipelined
        through the same TCP connection but they will be processed
        sequentially.
        See https://tools.ietf.org/html/draft-ietf-dnsop-5966bis-03
        Raises no exception: it's to be run in a thread

        :param addr: Tuple of the client's (IPv4 addr, Port) or
                     (IPv6 addr, Port, Flow info, Scope ID)
        :type addr: tuple
        :param client: Client socket
        :type client: socket.socket
        :raises: None
        """
        host, port = addr[:2]
        try:
            # The whole loop lives in a try/except block. On exceptions, the
            # connection is closed: there would be little chance to save
            # the connection after a struct error, a socket error.
            while True:
                # Decode the first 2 bytes containing the query length
                expected_length_raw = client.recv(2)
                if len(expected_length_raw) == 0:
                    break
                (expected_length,) = struct.unpack('!H', expected_length_raw)

                # Keep receiving data until we've got all the data we expect
                # The buffer contains only one query at a time
                buf = b''
                while len(buf) < expected_length:
                    recv_size = min(expected_length - len(buf),
                                    self._TCP_RECV_MAX_SIZE)
                    data = client.recv(recv_size)
                    if not data:
                        break
                    buf += data

                query = buf

                # Call into the DNS Application itself with payload and addr
                for response in self.app({'payload': query, 'addr': addr}):

                    # Send back a response only if present
                    if response is None:
                        continue

                    # Handle TCP Responses
                    msg_length = len(response)
                    tcp_response = struct.pack("!H", msg_length) + response
                    client.sendall(tcp_response)

        except socket.timeout:
            LOG.info(
                'TCP Timeout from: %(host)s:%(port)d',
                {
                    'host': host,
                    'port': port
                }
            )
        except OSError as e:
            errname = errno.errorcode[e.args[0]]
            LOG.warning(
                'Socket error %(err)s from: %(host)s:%(port)d',
                {
                    'host': host,
                    'port': port,
                    'err': errname
                }
            )
        except struct.error:
            LOG.warning(
                'Invalid packet from: %(host)s:%(port)d',
                {
                    'host': host,
                    'port': port
                }
            )
        except Exception:
            LOG.exception(
                'Unknown exception handling TCP request from: '
                '%(host)s:%(port)d',
                {
                    'host': host,
                    'port': port
                }
            )
        finally:
            client.close()

    def _dns_handle_udp(self, sock_udp):
        """Handle a DNS Query over UDP in a dedicated thread

        :param sock_udp: UDP socket
        :type sock_udp: socket.socket
        :raises: None
        """
        LOG.info('_handle_udp thread started')

        while self._running.is_set():
            addr = None
            try:
                # TODO(kiall): Determine the appropriate default value for
                #              UDP recvfrom.
                payload, addr = sock_udp.recvfrom(8192)

                LOG.debug(
                    'Handling UDP Request from: %(host)s:%(port)d',
                    {
                        'host': addr[0],
                        'port': addr[1]
                    }
                )

                # Dispatch a thread to handle the query
                self.tg.add_thread(self._dns_handle_udp_query, sock_udp, addr,
                                   payload)
            except socket.timeout:
                pass
            except OSError as e:
                errname = errno.errorcode[e.args[0]]
                addr = addr or (None, 0)
                LOG.warning(
                    'Socket error %(err)s from: %(host)s:%(port)d',
                    {
                        'host': addr[0],
                        'port': addr[1],
                        'err': errname
                    }
                )
            except Exception:
                addr = addr or (None, 0)
                LOG.exception(
                    'Unknown exception handling UDP request from: '
                    '%(host)s:%(port)d',
                    {
                        'host': addr[0],
                        'port': addr[1]
                    }
                )

    def _dns_handle_udp_query(self, sock, addr, payload):
        """
        Handle a DNS Query over UDP

        :param sock: UDP socket
        :type sock: socket.socket
        :param addr: Tuple of the client's (IP, Port)
        :type addr: tuple
        :param payload: Raw DNS query payload
        :type payload: string
        :raises: None
        """
        try:
            # Call into the DNS Application itself with the payload and addr
            for response in self.app({'payload': payload, 'addr': addr}):
                if response is not None:
                    sock.sendto(response, addr)
        except Exception:
            LOG.exception(
                'Unhandled exception while processing request from '
                '%(host)s:%(port)d',
                {
                    'host': addr[0],
                    'port': addr[1]
                }
            )


_launcher = None


def serve(server, workers=None):
    global _launcher
    if _launcher:
        raise RuntimeError(_('serve() can only be called once'))

    _launcher = service.launch(CONF, server, workers=workers,
                               restart_method='mutate')


def wait():
    try:
        _launcher.wait()
    except KeyboardInterrupt:
        LOG.debug('Caught KeyboardInterrupt, shutting down now')
    rpc.cleanup()
